/*

Copyright (C) 2003 Olivier VERLINDEN
    Service de Mecanique rationnelle, Dynamique et Vibrations
    Faculte Polytechnique de Mons
    31, Bd Dolez, 7000 MONS (Belgium)
    Olivier.Verlinden@fpms.ac.be

This file is part of EasyDyn

EasyDyn is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 2, or (at your option) any
later version.

EasyDyn is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with EasyDyn; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include <math.h>
#include <EasyDyn/mbs.h>


//---------------------------------------------------------------------------

void EasyDynmbsError(const char *texte)
{
 cout << "EasdyDynmbs ERROR: " << texte << "\n";
 exit(1);
}

//--------------------------------------------------------------------

void InitEasyDynmbs()

{
if (nbrdof==0) 
  EasyDynmbsError("nbrdof must be initialized before calling InitMbsSim");
if (application==0) 
  EasyDynmbsError("application must be initialized before calling InitMbsSim");
if (nbrbody==0) 
  EasyDynmbsError("nbrbody must be initialized before calling InitMbsSim");
InitEasyDynsim();
if (nbrdep>0) 
    {
    p=new double[nbrdep];
    pd=new double[nbrdep];
    pdd=new double[nbrdep];
    }
body=new structbody[nbrbody];
SetInertiaData();
// Let's verify the inertia data
int ibody;
for (ibody=0; ibody<nbrbody; ibody++)
   {
   body[ibody].vGpartial=new vec[nbrdof];
   body[ibody].omegapartial=new vec[nbrdof];
   body[ibody].vGrelpartial=new vec[nbrdof];
   body[ibody].omegarelpartial=new vec[nbrdof];
   body[ibody].underdof=new int[nbrdof];
   if (body[ibody].mass<=0) EasyDynmbsError("Mass null or negative");
   if (body[ibody].PhiG.Ixx<=0) EasyDynmbsError("Ixx null or negative");
   if (body[ibody].PhiG.Iyy<=0) EasyDynmbsError("Iyy null or negative");
   if (body[ibody].PhiG.Izz<=0) EasyDynmbsError("Izz null or negative");
   if (body[ibody].PhiG.Ixx>(body[ibody].PhiG.Iyy+body[ibody].PhiG.Izz)) 
      EasyDynmbsError("Ixx > Iyy+Izz");
   if (body[ibody].PhiG.Iyy>(body[ibody].PhiG.Ixx+body[ibody].PhiG.Izz)) 
      EasyDynmbsError("Iyy > Ixx+Izz");
   if (body[ibody].PhiG.Izz>(body[ibody].PhiG.Ixx+body[ibody].PhiG.Iyy)) 
      EasyDynmbsError("Izz > Ixx+Iyy");
   }
}

//---------------------------------------------------------------------------

void EndEasyDynmbs()

{
EndEasyDynsim();
delete body;
}

//--------------------------------------------------------------------

void ComputeInertiaEfforts()

{
int ibody;
vec LG;
// Calcul des reactions d'inertie
for (ibody=0; ibody<nbrbody; ibody++)
  {
  body[ibody].R=-body[ibody].mass*body[ibody].aG;
  // On calcule d'abord PHI*wd dans le repere local
  LG=body[ibody].PhiG*(body[ibody].T0G.R.inv()*body[ibody].omegad);
  body[ibody].MG=-(body[ibody].T0G.R*LG);
  LG=body[ibody].T0G.R*(body[ibody].PhiG*(body[ibody].T0G.R.inv()*body[ibody].omega));
  body[ibody].MG-=(body[ibody].omega^LG);
  }
}

//--------------------------------------------------------------------

void ComputePartialVelocitiesDefault()

{

ComputeMotion();

int ibody,idof;

for (ibody=0; ibody<nbrbody; ibody++) 
    { 
    body[ibody].vGsto=body[ibody].vG;
    body[ibody].omegasto=body[ibody].omega;
    }

for (idof=0; idof<nbrdof; idof++)
  {
  qd[idof]+=1.0;
  ComputeMotion();
  qd[idof]-=1.0;
  for (ibody=0; ibody<nbrbody; ibody++) 
    {
    body[ibody].vGpartial[idof]=body[ibody].vG-body[ibody].vGsto;
    body[ibody].omegapartial[idof]=body[ibody].omega-body[ibody].omegasto;
    }
  }

// Restore intial state
ComputeMotion();
for (ibody=0; ibody<nbrbody; ibody++) 
    { 
    body[ibody].vG=body[ibody].vGsto;
    body[ibody].omega=body[ibody].omegasto;
    }

}

//--------------------------------------------------------------------

void ComputeResidualmbs()

{
// Let's control the user assiduity
if (body==0) 
  EasyDynmbsError("Probably InitMbsSim was not called (body not allocated)");

ComputeMotion();
ComputePartialVelocities();
ComputeInertiaEfforts();
AddAppliedEfforts();

int ibody,idof;

for (idof=0; idof<nbrdof; idof++)
  {
  f[idof]=0;
  for (ibody=0; ibody<nbrbody; ibody++) 
    f[idof]-=body[ibody].R*body[ibody].vGpartial[idof]
	    +body[ibody].MG*body[ibody].omegapartial[idof];
  }
}

//--------------------------------------------------------------------

void ComposeMotion(int ibody, int ibodyref)
	
{
  if ((ibody<0) || (ibody>nbrbody))
   EasyDynmbsError("ComposeMotion, incorrect number for ibody");
  if ((ibodyref<0) || (ibodyref>nbrbody)) 
   EasyDynmbsError("ComposeMotion, incorrect number for ibodyref");
  if (ibodyref>ibody) 
   EasyDynmbsError("ComposeMotion, ibody must be larger than ibodyref");

// Let's first project all relative terms in the absolute coordinate system
vec erel=body[ibodyref].T0G.R*body[ibody].TrefG.e,
    vrel=body[ibodyref].T0G.R*body[ibody].vGrel,
    arel=body[ibodyref].T0G.R*body[ibody].aGrel,
    wrel=body[ibodyref].T0G.R*body[ibody].omegarel,
    wdrel=body[ibodyref].T0G.R*body[ibody].omegadrel;

// Let's go for the composition
body[ibody].T0G=body[ibodyref].T0G*body[ibody].TrefG;
body[ibody].vG=body[ibodyref].vG+(body[ibodyref].omega^erel)+vrel;
body[ibody].omega=body[ibodyref].omega+wrel;
body[ibody].aG=body[ibodyref].aG+(body[ibodyref].omegad^erel)
             +(body[ibodyref].omega^(body[ibodyref].omega^erel))
             +2*(body[ibodyref].omega^vrel)+arel;
body[ibody].omegad=body[ibodyref].omegad+(body[ibodyref].omega^wrel)
                  +wdrel;
}

//--------------------------------------------------------------------

void ComposePartialVelocities(int ibody, int ibodyref)
	
{
  if ((ibody<0) || (ibody>nbrbody))
   EasyDynmbsError("ComposePartialVelocities, incorrect number for ibody");
  if ((ibodyref<0) || (ibodyref>nbrbody)) 
   EasyDynmbsError("ComposePartialVelocities, incorrect number for ibodyref");
  if (ibodyref>ibody) 
   EasyDynmbsError("ComposePartialVelocities, ibody must be larger than ibodyref");

// Let's first project all relative terms in the absolute coordinate system
vec erel=body[ibodyref].T0G.R*body[ibody].TrefG.e;
vec vrel, wrel;
int idof;
for (idof=0; idof<nbrdof; idof++)
  {
  vrel=body[ibodyref].T0G.R*body[ibody].vGrelpartial[idof];
  wrel=body[ibodyref].T0G.R*body[ibody].omegarelpartial[idof];
  body[ibody].vGpartial[idof]=body[ibodyref].vGpartial[idof]
                             +(body[ibodyref].omegapartial[idof]^erel)
                             +vrel;
  body[ibody].omegapartial[idof]=body[ibodyref].omegapartial[idof]+wrel;
  }
}

//--------------------------------------------------------------------

void AddGravityForces(vec grav)
	
{
int ibody;
for (ibody=0; ibody<nbrbody; ibody++) body[ibody].R+=body[ibody].mass*grav;
}
	
//--------------------------------------------------------------------

void AddSpringForce(double K, double L0, 
                    int ibodyA, vec rA, int ibodyB, vec rB)

{
vec uAB=body[ibodyB].T0G*rB-body[ibodyA].T0G*rA;
double L=uAB.length();
uAB.unite();
vec force=K*(L-L0)*uAB;
body[ibodyA].R+=force;
body[ibodyA].MG+=((body[ibodyA].T0G.R*rA)^force);
body[ibodyB].R-=force;
body[ibodyB].MG-=((body[ibodyB].T0G.R*rB)^force);
}

//--------------------------------------------------------------------

void AddDamperForce(double C,
                    int ibodyA, vec rA, int ibodyB, vec rB)

{
vec uAB=body[ibodyB].T0G*rB-body[ibodyA].T0G*rA;
uAB.unite();
vec vrel=body[ibodyB].vG+(body[ibodyB].omega^(body[ibodyB].T0G.R*rB))
         -body[ibodyA].vG-(body[ibodyA].omega^(body[ibodyA].T0G.R*rA));
vec force=C*(vrel*uAB)*uAB;
body[ibodyA].R+=force;
body[ibodyA].MG+=((body[ibodyA].T0G.R*rA)^force);
body[ibodyB].R-=force;
body[ibodyB].MG-=((body[ibodyB].T0G.R*rB)^force);
}

//--------------------------------------------------------------------

void AddTyreEfforts(int ibody, vec axe, structtyre tyre)
{
// Determination of the conventional contact coordinate system
// z perpendicular to the ground
// y projection of the rotation axis on the ground
mth tSAE;
vec zSAE(0,0,1), uaxe=body[ibody].T0G.R*axe;
uaxe.unite();
vec xSAE=uaxe^zSAE;
xSAE.unite();
vec ySAE=zSAE^xSAE;
tSAE.R.setux(xSAE);
tSAE.R.setuy(ySAE);
tSAE.R.setuz(zSAE);
tSAE.e=body[ibody].T0G.e+(tyre.r1-tyre.r2)*(uaxe^xSAE)-tyre.r2*zSAE;

// Determination of penetration
double pen=-tSAE.e.z;

if (pen<=0) return;  // all efforts are of course equal to zero

// The SAE coordinate system is put on the ground
tSAE.e.z=0;
// Determination of contact point velocity
vec vP=body[ibody].vG+(body[ibody].omega^(tSAE.e-body[ibody].T0G.e));
// Determination of relevant coordinates
double vPxSAE=vP*xSAE, vPySAE=vP*ySAE, vPzSAE=vP*zSAE,
       vCxSAE=body[ibody].vG*xSAE;
// Determination of reference velocity
double vref;
if (vCxSAE*vPxSAE>0) vref=fabs(vCxSAE);  /* braking */
    else vref=fabs(vCxSAE-vPxSAE); /* traction */
if (vref<1E-6) vref=1E-6;
// Determination of slips
double slong=vPxSAE/vref, slat=vPySAE/vref, scamber=uaxe*zSAE;
// Determination of normal force
double Fz=pen*tyre.Kz-sqrt(pen*tyre.Kz/tyre.Fznom)*tyre.Cz*vPzSAE;

if (Fz<0) return;

double lc=sqrt(8*tyre.r1*pen);
// Determination of tyre efforts according to the model developed at
// the University of Arizona by Gim Gwanghun (cf Ph.D and doc ADAMS)

double Clong,Clat,Ccamber,Stot,B1,B2,B3,fClb,Sn,ln;
Stot=sqrt(slong*slong+slat*slat);
 if (Stot<1E-6) Stot=1E-6;
// Determination of friction coefficient according to total slip
double Stot2=(Stot<1 ? Stot : 1);
fClb=tyre.fClbs*(1-Stot2)+tyre.fClbd*Stot2;
// Determination of stiffness coefficients in terms of vertical load
Clong=tyre.Clongnom*pow((Fz/tyre.Fznom),tyre.nlong);
Clat=tyre.Clatnom*pow((Fz/tyre.Fznom),tyre.nlat);
Ccamber=tyre.Ccambernom*pow((Fz/tyre.Fznom),tyre.ncamber);
// Is there skid ?
int skid=((Clong*Clong*slong*slong
   +(Clat*slat+3*Ccamber*scamber)*(Clat*slat+3*Ccamber*scamber))
	       > (9*fClb*fClb*Fz*Fz));
// Determination of efforts
double Fx=0,Fy=0,My=0,Mz=0;
if (skid) // in case of skid
      {
	// cout << "skid\n";
      Fx=-fClb*slong*Fz/Stot;
      Fy=-fClb*slat*Fz/Stot;
      My=0.0;
      Mz=-0.6*fClb*slong*fClb*slat*Fz*Fz*lc/(Stot*Stot*Clat);
      }
// Computation of non dimensional length of slip area
else // no skid
      {
      B1=9*(fClb*fClb*Fz*Fz-Ccamber*Ccamber*scamber*scamber);
      B2=3*Clat*slat*Ccamber*scamber;
      B3=-(Clat*Clat*slat*slat+Clong*Clong*slong*slong);
      Sn=(B2+sqrt(B2*B2-B1*B3))/B1; ln=1-Sn;
      // Calcul des efforts dans le cas lineaire des faibles glissements
        /*  if (Sn<1E-6)
        {
        cout << "\nPetits glissements Sn=" << Sn << endl;
        Eff.Fx=-Clong*slong;
        Eff.Fy=-Clat*slat;
        Eff.My=0.0;
        Eff.Mz=-2*Clong*slong*slat*lc/3+Clat*slat*lc/6;
        return Eff;
        }*/
      // Computation of contact efforts in general
      Fx=-Clong*slong*ln*ln-fClb*slong*Fz*(1+ln*ln*(2*ln-3))/Stot;
      Fy=-Clat*slat*ln*ln
	     -fClb*slat*Fz*(1+ln*ln*(2*ln-3))/Stot
	     -Ccamber*scamber*ln*ln*(3-2*ln);
      My=0.0;
      Mz=-2*Clong*slong*slat*lc*ln*ln*ln/3 
         -0.6*fClb*slong*fClb*slat*Fz*Fz*
	  lc*(1+ln*ln*ln*(-10+15*ln-6*ln*ln))/(Stot*Stot*Clat);
      // Differentiation of case 3 from cases 1 and 2 (cf. doc. ADAMS)
      if((Clat*Clat*slat*slat+Ccamber*scamber*Clat*slat)>0)
        {
        Mz+=(Clat*slat*(-0.5+2*ln/3)
	     +1.5*(fClb*slat*Fz/Stot-Ccamber*scamber)*Sn*Sn)*lc*ln*ln;
        }
      else
        {
        Mz+=Clat*slat*lc*ln/6;
        }
      }
vec Ftot=Fx*xSAE+Fy*ySAE+Fz*zSAE;
body[ibody].R+=Ftot;
 body[ibody].MG+=My*ySAE+Mz*zSAE+((tSAE.e-body[ibody].T0G.e)^Ftot);
}

//--------------------------------------------------------------------

void AddContactPlanePointForce(double Kn, double pk, 
                               double Cdamp, double pd,
                               double ffrict, double vglim,
                               int ibodyplane, vec rplane, vec nplane,
		               int ibodypoint, vec rpoint)

{
vec eplane, vplane, epoint, vpoint, nglob;

eplane=body[ibodyplane].T0G*rplane;
epoint=body[ibodypoint].T0G*rpoint;
nglob=body[ibodyplane].T0G.R*nplane;
nglob.unite();
double pen=-(epoint-eplane)*nglob;
if (pen>0)
 {
 vec eforce=epoint+0.5*pen*nglob;
 vplane=body[ibodyplane].vG
      +(body[ibodyplane].omega^(eforce-body[ibodyplane].T0G.e));
 vpoint=body[ibodypoint].vG
      +(body[ibodypoint].omega^(eforce-body[ibodypoint].T0G.e));
 vec vr=(vpoint-vplane);
 double pend=-vr*nglob;
 vec vt=vr+pend*nglob;
 double Fn=Kn*pow(pen,pk)+Cdamp*pow(pen,pd)*pend;
 if (Fn<0) Fn=0;
 vec force=Fn*nglob;
 if (vt.length()>vglim) force-=ffrict*Fn*vt/vt.length();
 else force-=ffrict*Fn*vt/vglim;
 body[ibodypoint].R+=force;
 body[ibodypoint].MG+=((eforce-body[ibodypoint].T0G.e)^force);
 body[ibodyplane].R-=force;
 body[ibodyplane].MG-=((eforce-body[ibodyplane].T0G.e)^force);
 }
}

//--------------------------------------------------------------------

void CreateVmoFile(scene &sc)
	
{
// Storage of mth in reference configuration
ComputeMotion();
mth *T0Gref;
T0Gref=new mth[nbrbody];
int ibody;
for (ibody=0; ibody<nbrbody; ibody++) T0Gref[ibody]=body[ibody].T0G;

// Resaving the reference configuration to be sure
char *FileName;
FileName=new char[strlen(application)+5];
strcpy(FileName,application);
strcat(FileName,".vol");
sc.CreateVolFile(FileName);

// Opening the mod file
strcpy(FileName,application);
strcat(FileName,".mod");
ifstream ModFile(FileName);

// Opening the vmo file
strcpy(FileName,application);
strcat(FileName,".vmo");
ofstream VmoFile(FileName);

double alpha,omega,freq,amor,*qref,*qreal,*qimag,coeff,angle,anglemax;
int iddl,itmp;
trot Rrel;

// Storing the reference position
qref=new double[nbrdof];
for (iddl=0; iddl<nbrdof; iddl++) qref[iddl]=q[iddl];
qreal=new double[nbrdof];
qimag=new double[nbrdof];
while (!ModFile.rdstate())
  {
  ModFile >> itmp >> alpha >> omega >> freq >> amor;
  coeff=0;
  // Load the mode and determine the amplitude
  for (iddl=0; iddl<nbrdof; iddl++) if (!ModFile.rdstate())
      {
      ModFile >> qreal[iddl] >> qimag[iddl];
      coeff+=qreal[iddl]*qreal[iddl]+qimag[iddl]*qimag[iddl];
      }
  if (!ModFile.rdstate()) // data treated only if no problem after reading
    {
    // Normalize the amplitude of the mode
    for (iddl=0; iddl<nbrdof; iddl++) 
      {
      qreal[iddl]*=0.01/sqrt(coeff);
      qimag[iddl]*=0.01/sqrt(coeff);
      } 
    // Determine the maximum angle for real part of mode
    for (iddl=0; iddl<nbrdof; iddl++) q[iddl]=qref[iddl]+qreal[iddl];
    ComputeMotion();
    anglemax=0;
    for (ibody=0; ibody<nbrbody; ibody++) 
      {
      Rrel=T0Gref[ibody].R.inv()*body[ibody].T0G.R;
      angle=2*acos(0.5*sqrt(1+Rrel.r11+Rrel.r22+Rrel.r33));
      if (angle>anglemax) anglemax=angle;
      } 
    // and go on for imaginary part of mode
    for (iddl=0; iddl<nbrdof; iddl++) q[iddl]=qref[iddl]+qimag[iddl];
    ComputeMotion();
    for (ibody=0; ibody<nbrbody; ibody++) 
      {
      Rrel=T0Gref[ibody].R.inv()*body[ibody].T0G.R;
      angle=2*acos(0.5*sqrt(1+Rrel.r11+Rrel.r22+Rrel.r33));
      if (angle>anglemax) anglemax=angle;
      } 
    // Adapt the amplitude of the mode
    for (iddl=0; iddl<nbrdof; iddl++) 
      {
      qreal[iddl]*=0.3/anglemax;
      qimag[iddl]*=0.3/anglemax;
      } 
    // Store mode in VmoFile
    VmoFile << alpha << " " << omega << " " << freq << " " << amor << endl;
    for (iddl=0; iddl<nbrdof; iddl++) q[iddl]=qref[iddl]+qreal[iddl];
    ComputeMotion();
    sc.WriteCoord(VmoFile); 
    for (iddl=0; iddl<nbrdof; iddl++) q[iddl]=qref[iddl]+qimag[iddl];
    ComputeMotion();
    sc.WriteCoord(VmoFile); 
    }
  }
ModFile.close();
VmoFile.close();

// restore initial position
for (iddl=0; iddl<nbrdof; iddl++) q[iddl]=qref[iddl];
ComputeMotion();
	
delete T0Gref;
delete qref;
delete qreal;
delete qimag;
delete FileName;

}

//--------------------------------------------------------------------
